"""
The omc3-gui library
~~~~~~~~~~~~~~~~~~~~

omc3-gui provides graphical user interfaces for the ``omc3`` package for the optics measurements and corrections group (OMC) at CERN.

:copyright: pyLHC/OMC-Team working group.
"""

__title__ = "omc3-gui"
__description__ = "QT Graphical User Interface wrapper of the ``omc3`` package"
__url__ = "https://github.com/pylhc/omc3_gui"
__version__ = "1.0.1"
__author__ = "pylhc"
__author_email__ = "pylhc@github.com"

__all__ = [__version__]
